// font.cc

#include <iostream>
#include <list>
#include <map>
#include <string>
#include <vector>

#define GLEW_STATIC 1
#include <GL/glew.h>
#include <GL/freeglut.h>

#include "gl-utils.h"
#include "shader.h"

#include "font-manager.h"

const unsigned FONT_TEXTURE_SIZE = 512;
const int FONT_TEXTURE_PAD = 3;
const int TEXT_MAX_SIZE = 1024; // How many characters can be output by one drawcall.

/*****************************************************************************\
 * FontManager::FontFaceId::FontFaceId                                       *
\*****************************************************************************/
FontManager::FontFaceId::FontFaceId(const std::string &name, int size, bool serial)
  : font_name(name), font_size(size), from_serial(serial)
{ }

/*****************************************************************************\
 * FontManager::FontFace::FontFace                                           *
\*****************************************************************************/
FontManager::FontFace::FontFace(FT_Library &library, const std::string &filename, int size, int ppi) : freetype_library(library)
{
  FT_Error error;
  error = FT_New_Face(freetype_library, filename.c_str(), 0, &freetype_face);
  if (error)
  {
    std::cerr << "Cannot load the '" << filename << "' font file.\n";
    return;
  }
  error = FT_Select_Charmap(freetype_face, FT_ENCODING_UNICODE);
  if (error)
  {
    std::cerr << "Cannot use the unicode encoding with the '" << filename << "' font file.\n";
  }
  FT_Set_Char_Size(freetype_face, size << 6, size << 6, ppi, ppi); // FIXME, test error code

  GLuint texture;
  glGenTextures(1, &texture);
  glBindTexture(GL_TEXTURE_2D, texture);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  std::vector<unsigned char> init_tex;
  for (unsigned i = 0; i < FONT_TEXTURE_SIZE * FONT_TEXTURE_SIZE; i++)
    init_tex.push_back(0);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RED, FONT_TEXTURE_SIZE, FONT_TEXTURE_SIZE, 0, GL_RED, GL_UNSIGNED_BYTE, &init_tex[0]);
  textures.push_back(texture);
  x_off = y_off = y_max = 0;
}

/*****************************************************************************\
 * FontManager::FontManager                                                  *
\*****************************************************************************/
FontManager::FontManager()
{
  FT_Error error;
  error = FT_Init_FreeType(&freetype_library);
  if (error)
  {
    std::cerr << "Cannot create a freetype library handle.\n";
    ::exit(-1);
  }
}

/*****************************************************************************\
 * FontManager::~FontManager                                                 *
\*****************************************************************************/
FontManager::~FontManager()
{
  FT_Error error;
  error = FT_Done_FreeType(freetype_library);
  if (error)
  {
//    DDX_LOG_ERROR("DDX-Widget_FontManager", "Cannot destroy a freetype library handle");
  }
//  DDX_LOG_DEBUG("DDX-Widget_FontManager", "Destructor");
}

/*---------------------------------------------------------------------------*\
 * FontManager::init_common                                                  *
\*---------------------------------------------------------------------------*/
void FontManager::init_common()
{
  font_faces.clear();
  std::string vertex_text_shader =
"#version 330\n"
"\n"
"uniform vec2 size = vec2(800.0, 600.0);\n"
"\n"
"in vec2 position;\n"
"in vec2 tex_coord;\n"
"\n"
"out vec2 f_tex_coord;\n"
"\n"
"void main(void)\n"
"{\n"
"  gl_Position = vec4(position.x / size.x * 2.0 - 1.0,\n"
"                     1.0 - position.y / size.y * 2.0,\n"
"                      0.0, 1.0);\n"
"  f_tex_coord = tex_coord;\n"
"}\n";

  std::string fragment_text_shader =
"#version 330\n"
"\n"
"uniform sampler2D text_sampler;\n"
"uniform vec4 color = vec4(1.0, 0.0, 1.0, 1.0);\n"
"\n"
"in vec2 f_tex_coord;\n"
"\n"
"out vec4 out_color;\n"
"\n"
"void main(void)\n"
"{\n"
"  out_color = vec4(color.rgb, color.a * texture(text_sampler, f_tex_coord).r);\n"
"  //out_color = vec4(1.0, 1.0, 0.0, 1.0);\n"
"}\n";
  // Creating the text rendering program
  std::vector<std::string> attributes;
  attributes.push_back("position");
  attributes.push_back("tex_coord");
  program = new Shader();
  program->attach_vertex(vertex_text_shader);
  program->attach_fragment(fragment_text_shader);
  program->set_attributes(attributes);
  program->link();
  program->activate();
  program->set("text_sampler", 0);

  // create VAO, VBO, IBO...
  glGenVertexArrays(1, &vao);
  glBindVertexArray(vao);

  glGenBuffers(1, &vbo);
  glBindBuffer(GL_ARRAY_BUFFER, vbo);
  glBufferData(GL_ARRAY_BUFFER, TEXT_MAX_SIZE * 2 * 4 * sizeof(vec2), 0, GL_DYNAMIC_DRAW);
  glEnableVertexAttribArray(0);
  glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, 2 * sizeof(vec2), nullptr);
  glEnableVertexAttribArray(1);
  glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 2 * sizeof(vec2), BUFFER_OFFSET(sizeof(vec2)));
  glGenBuffers(1, &vbo_indices);
  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vbo_indices);
  std::vector<GLushort> indices;
  for (int i = 0; i < TEXT_MAX_SIZE; i++)
  {
    indices.push_back(4 * i + 0);
    indices.push_back(4 * i + 1);
    indices.push_back(4 * i + 2);
    indices.push_back(4 * i + 1);
    indices.push_back(4 * i + 2);
    indices.push_back(4 * i + 3);
  }
  glBufferData(GL_ELEMENT_ARRAY_BUFFER, TEXT_MAX_SIZE * 6 * sizeof(GLushort), &indices[0], GL_STATIC_DRAW);
  glBindVertexArray(0);
  glBindBuffer(GL_ARRAY_BUFFER, 0);
  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

  current_ppi = 100;
}

/*****************************************************************************\
 * FontManager::init                                                         *
\*****************************************************************************/
void FontManager::init()
{
  init_common();
  set_font("fonts/luxisr.ttf", 10);
}

/*****************************************************************************\
 * FontManager::exit                                                         *
\*****************************************************************************/
void FontManager::exit()
{

}

/*****************************************************************************\
 * FontManager::set_color                                                    *
\*****************************************************************************/
void FontManager::set_color(const vec4 &color)
{
  program->activate();
  program->set("color", color);
}

/*****************************************************************************\
 * FontManager::set_screen_size                                              *
\*****************************************************************************/
void FontManager::set_screen_size(const vec2 &size)
{
  program->activate();
  program->set("size", size);
  glUseProgram(0);
}

/*****************************************************************************\
 * FontManager::set_font                                                     *
\*****************************************************************************/
void FontManager::set_font(const std::string &filename, int size)
{
  FontFaceId id(filename, size);

  current_font_name = filename;
  current_font_size = size;
  std::map<FontFaceId,FontFace*>::iterator it;
  it = font_faces.find(id);
  if (it != font_faces.end())
    current_font_face = it->second;
  else
  {
    current_font_face = new FontFace(freetype_library, filename, size, current_ppi);
    font_faces[id] = current_font_face;
  }
}

/*****************************************************************************\
 * FontManager::set_font_size                                                *
\*****************************************************************************/
void FontManager::set_font_size(int size)
{
  set_font(current_font_name, size);
}

/*===========================================================================*\
 *                   get_next_utf8                                           *
\*===========================================================================*/
static int get_next_utf8(const char *&buf)
{
  unsigned char d = *buf++, d2, d3, d4;

  if (d < 0x80)
    return d;
  if ((d & 0xe0) == 0xc0) // 0b1100 0b1101
  {
    d2 = *buf++;
    if ((d2 & 0xc0) != 0x80)
      return 0;
    return ((d & 0x1f) << 6) | (d2 & 0x3f);
  }
  if ((d & 0xf0) == 0xe0) // Ob1110
  {
    d2 = *buf++;
    d3 = *buf++;
    if ((d2 & 0xc0) != 0x80 || (d3 & 0xc0) != 0x80)
      return 0;
    return ((d & 0x0f) << 12) | ((d2 & 0x3f) << 6) | (d3 & 0x3f);
  }
  d2 = *buf++;
  d3 = *buf++;
  d4 = *buf++;
  if ((d2 & 0xc0) != 0x80 || (d3 & 0xc0) != 0x80 || (d4 & 0xc0) != 0x80)
    return 0;
  return ((d & 0x0f) << 18) | ((d2 & 0x3f) << 12) | ((d3 & 0x3f) << 6) | (d4 & 0x3f);
}

/*****************************************************************************\
 * FontManager::FontFace::get_glyph                                          *
\*****************************************************************************/
FontManager::FontFace::Glyph FontManager::FontFace::get_glyph(const char *&text)
{
  int            index;
  unsigned char *src;
  //FT_Error error;
  Glyph glyph;
  char letter = *text;

  index = FT_Get_Char_Index(freetype_face, get_next_utf8(text));
  if (glyphes.find(index) != glyphes.end())
    return glyphes[index];
  /*error = */FT_Load_Glyph(freetype_face, index, 0);
  if (freetype_face->glyph->format != FT_GLYPH_FORMAT_BITMAP) {
    /*error = */FT_Render_Glyph(freetype_face->glyph, FT_RENDER_MODE_NORMAL);
  }

  glyph.width = freetype_face->glyph->bitmap.width;
  glyph.height = freetype_face->glyph->bitmap.rows;

  if (glyph.width >= int(FONT_TEXTURE_SIZE) || glyph.height >= int(FONT_TEXTURE_SIZE))
  {
    std::cerr << "Enormous glyph! >" << letter << "< is " << glyph.width << " by " << glyph.height << "!\n";
  }

  if (x_off + glyph.width >= FONT_TEXTURE_SIZE)
  {
    x_off = 0;
    y_off += y_max + FONT_TEXTURE_PAD;
    y_max = 0;
  }
  if (y_off + glyph.height >= FONT_TEXTURE_SIZE)
  {
    GLuint texture;
    glGenTextures(1, &texture);
    glBindTexture(GL_TEXTURE_2D, texture);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    std::vector<unsigned char> init_tex;
    for (unsigned i = 0; i < FONT_TEXTURE_SIZE * FONT_TEXTURE_SIZE; i++)
      init_tex.push_back(0);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RED, FONT_TEXTURE_SIZE, FONT_TEXTURE_SIZE, 0, GL_RED, GL_UNSIGNED_BYTE, &init_tex[0]);
    textures.push_back(texture);
    x_off = y_off = y_max = 0;
    aux_buffers.resize(textures.size() - 1);
  }
  src = freetype_face->glyph->bitmap.buffer;
  glyph.data = new unsigned char[glyph.width * glyph.height];
  memcpy(glyph.data, src, glyph.width * glyph.height);

//  glPushClientAttrib(GL_CLIENT_PIXEL_STORE_BIT);
  glPixelStorei(GL_UNPACK_LSB_FIRST, GL_FALSE);
  glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glBindTexture(GL_TEXTURE_2D, textures[textures.size() - 1]);
  //DDX_LOG_INFO("DDX-Widget_FontManager", "x_off: %d, y_off: %d", x_off, y_off);
  glTexSubImage2D(GL_TEXTURE_2D, 0, x_off, y_off, glyph.width, glyph.height, GL_RED, GL_UNSIGNED_BYTE, src);
//  glPopClientAttrib();

  glyph.uv[0] = vec2(x_off,               y_off) / float(FONT_TEXTURE_SIZE);
  glyph.uv[1] = vec2(x_off + glyph.width, y_off) / float(FONT_TEXTURE_SIZE);
  glyph.uv[2] = vec2(x_off,               y_off + glyph.height) / float(FONT_TEXTURE_SIZE);
  glyph.uv[3] = vec2(x_off + glyph.width, y_off + glyph.height) / float(FONT_TEXTURE_SIZE);

  y_max = std::max(y_max, glyph.height);
  glyph.left       = freetype_face->glyph->bitmap_left;
  glyph.top        = freetype_face->glyph->bitmap_top;
  glyph.advance_x  = freetype_face->glyph->advance.x;
  glyph.advance_y  = freetype_face->glyph->advance.y;
  glyph.texture_index = textures.size() - 1;
  glyph.index = index;

  x_off += glyph.width + FONT_TEXTURE_PAD;

  glyphes[index] = glyph;

  return glyph;
}

/*****************************************************************************\
 * FontManager::get_glyph_face                                               *
\*****************************************************************************/
void FontManager::get_glyph_face(const char *text, int &width, int &height, const unsigned char *&data)
{
  current_font_face->get_glyph_face(text, width, height, data);
}

/*****************************************************************************\
 * FontManager::FontFace::get_glyph_face                                     *
\*****************************************************************************/
void FontManager::FontFace::get_glyph_face(const char *text, int &width, int &height, const unsigned char *&data)
{
  Glyph glyph = get_glyph(text);
  width = glyph.width;
  height = glyph.height;
  data = glyph.data;
}

/*****************************************************************************\
 * FontManager::draw_text                                                    *
\*****************************************************************************/
void FontManager::draw_text(int x, int y, const char *text)
{
  glActiveTexture(GL_TEXTURE0);
  glBindVertexArray(vao);
  glBindBuffer(GL_ARRAY_BUFFER, vbo);
  program->activate();

  current_font_face->draw_text(x, y, text);

  glBindVertexArray(0);
  glBindBuffer(GL_ARRAY_BUFFER, 0);
}

/*****************************************************************************\
 * FontManager::FontFace::draw_text                             *
\*****************************************************************************/
void FontManager::FontFace::draw_text(int x, int y, const char *text)
{
  unsigned i = 0;
  //vec2 *buffer = reinterpret_cast<vec2*>(glMapBuffer(GL_ARRAY_BUFFER, GL_WRITE_ONLY)); PRINT_OPENGL_ERROR();
  std::vector<vec2> buffer;

  aux_buffers.clear();
  aux_buffers.resize(textures.size() - 1);

  int previous = 0;
  bool use_kerning = FT_HAS_KERNING(freetype_face);

  while (*text)
  {
    Glyph glyph = get_glyph(text);
    if (use_kerning && previous && glyph.index)
    {
      FT_Vector delta;
      FT_Get_Kerning(freetype_face, previous, glyph.index, FT_KERNING_DEFAULT, &delta);
      x += delta.x / 64;
    }
    if (glyph.texture_index == 0)
    {
      // buffer[8 * i + 0] = vec2(x, y - glyph.top);
      // buffer[8 * i + 1] = glyph.uv[0];
      // buffer[8 * i + 2] = vec2(x + glyph.width, y - glyph.top);
      // buffer[8 * i + 3] = glyph.uv[1];
      // buffer[8 * i + 4] = vec2(x, y - glyph.top + glyph.height);
      // buffer[8 * i + 5] = glyph.uv[2];
      // buffer[8 * i + 6] = vec2(x + glyph.width, y - glyph.top + glyph.height);
      // buffer[8 * i + 7] = glyph.uv[3];
      buffer.push_back(vec2(x, y - glyph.top));
      buffer.push_back(glyph.uv[0]);
      buffer.push_back(vec2(x + glyph.width, y - glyph.top));
      buffer.push_back(glyph.uv[1]);
      buffer.push_back(vec2(x, y - glyph.top + glyph.height));
      buffer.push_back(glyph.uv[2]);
      buffer.push_back(vec2(x + glyph.width, y - glyph.top + glyph.height));
      buffer.push_back(glyph.uv[3]);
      i++;
    }
    else
    {
      aux_buffers[glyph.texture_index - 1].push_back(vec2(x, y - glyph.top));
      aux_buffers[glyph.texture_index - 1].push_back(glyph.uv[0]);
      aux_buffers[glyph.texture_index - 1].push_back(vec2(x + glyph.width, y - glyph.top));
      aux_buffers[glyph.texture_index - 1].push_back(glyph.uv[1]);
      aux_buffers[glyph.texture_index - 1].push_back(vec2(x, y - glyph.top + glyph.height));
      aux_buffers[glyph.texture_index - 1].push_back(glyph.uv[2]);
      aux_buffers[glyph.texture_index - 1].push_back(vec2(x + glyph.width, y - glyph.top + glyph.height));
      aux_buffers[glyph.texture_index - 1].push_back(glyph.uv[3]);
    }
    x += glyph.advance_x / 64;
    previous = glyph.index;
  }
  //  glUnmapBuffer(GL_ARRAY_BUFFER); PRINT_OPENGL_ERROR();
  glBindTexture(GL_TEXTURE_2D, textures[0]);
  glBufferSubData(GL_ARRAY_BUFFER, 0, buffer.size() * sizeof(vec2), &buffer[0]);
  glDrawElements(GL_TRIANGLES, 6 * i, GL_UNSIGNED_SHORT, 0);

  for (unsigned i = 1; i <= aux_buffers.size(); i++)
    if (aux_buffers[i - 1].size())
    {
      glBindTexture(GL_TEXTURE_2D, textures[i]);
      glBufferSubData(GL_ARRAY_BUFFER, 0, aux_buffers[i - 1].size() * sizeof(vec2), &aux_buffers[i - 1][0]);
      glDrawElements(GL_TRIANGLES, 6 * aux_buffers[i - 1].size() / 8, GL_UNSIGNED_SHORT, 0);
    }
}

/*****************************************************************************\
 * FontManager::get_text_size                                                *
\*****************************************************************************/
void FontManager::get_text_size(const char *text,
                                int &string_width, int &string_height,
                                int &string_ascent)
{
  current_font_face->get_text_size(text, current_font_size, current_ppi,
                                   string_width, string_height, string_ascent);
}

/*****************************************************************************\
 * FontManager::FontFace::get_text_size                                      *
\*****************************************************************************/
void FontManager::FontFace::get_text_size(const char *text, int font_height, int ppi,
                                          int &string_width, int &string_height,
                                          int &string_ascent)
{
  FT_Error error;
  int index;
  FT_Vector pen;
  int ascent, descent;

  FT_Set_Char_Size(freetype_face, font_height << 6, font_height << 6, ppi, ppi);

  ascent = descent = 0;
  pen.x = pen.y = 0;

  int previous = 0;
  bool use_kerning = FT_HAS_KERNING(freetype_face);

  while (*text)
  {
    FT_Set_Transform(freetype_face, 0, &pen);

    index = FT_Get_Char_Index(freetype_face, get_next_utf8(text));
    if (use_kerning && previous && index)
    {
      FT_Vector delta;
      FT_Get_Kerning(freetype_face, previous, index, FT_KERNING_DEFAULT, &delta);
      pen.x += delta.x;
    }

    error = FT_Load_Glyph(freetype_face, index, 0);
    if (error)
    {
      std::cerr << "Failed to load glyph for character '"<< *text << "'\n";
      break;
    }

    if (freetype_face->glyph->format != FT_GLYPH_FORMAT_BITMAP)
    {
      error = FT_Render_Glyph(freetype_face->glyph, FT_RENDER_MODE_NORMAL);
      if (error)
      {
        std::cerr << "Failed to render glyph for character '" << *text << "'\n";
        break;
      }
    }
    ascent = std::max(ascent, freetype_face->glyph->bitmap_top);
    descent = std::max(descent, int(freetype_face->glyph->bitmap.rows -
        freetype_face->glyph->bitmap_top));

    pen.x += freetype_face->glyph->advance.x;
    pen.y += freetype_face->glyph->advance.y;
    previous = index;
  }
  string_width = pen.x >> 6;
  string_height = ascent + descent;
  string_ascent = ascent;
}
