//=================================================================================================================
//! \file      ddx-base_singleton.h
//! \brief     Our singleton pattern implementation.
///
//! \author    David Odin <david.odin@forma3dev.fr>
//! \date      Created on 2009-07-20
///
//! Modifications: $Id: ddx-base_singleton.h 765 2012-05-06 23:52:35Z dindinx $
///
//! \note (c) Forma3Dev - forma3dev.fr
//=================================================================================================================

#ifndef DDX_BASE__SINGLETON_H
#define DDX_BASE__SINGLETON_H

#include <cstdlib>
#ifndef ANDROID
#include <stdexcept>
#endif
///////////////////////////////////////////////////////////////////////
// Policies
///////////////////////////////////////////////////////////////////////
template <class T> struct CreateUsingNew
{
  static T*   create  ()     { return new T; }
  static void destroy (T* p) { delete p; }
};

template <class T> struct CreateStatic
{
  union MaxAlign
    {
      char t_[sizeof(T)];
      short int shortInt_;
      int int_;
      long int longInt_;
      float float_;
      double double_;
      long double longDouble_;
      struct Test;
      int Test::* pMember_;
      int (Test::*pMemberFn_)(int);
    };
  static T* create()
    {
      static MaxAlign staticMemory_;
      return new(&staticMemory_) T;
    }
  static void destroy(T* p)
    {
      p->~T();
    }
};

typedef void (*atexit_pfn_t)();
template <class T> struct DefaultLifetime
{
  static void schedule_destruction (T*, atexit_pfn_t pFun) { std::atexit (pFun); }
  static void on_dead_reference    ()
    {
#ifndef ANDROID
      throw std::logic_error("Dead Reference Detected");
#endif
    }
};

template <class T> class DeletableSingleton
{
public:
  static void schedule_destruction(T*, atexit_pfn_t pFun)
  {
    static bool first_pass = true;
    is_dead = false;
    deleter = pFun;
    if (first_pass || need_callback)
      {
        std::atexit (atexit_callback);
        first_pass = false;
        need_callback = false;
      }
  }

  static void on_dead_reference() { }

  ///  delete singleton object manually
  static void graceful_delete ()
  {
    if (is_dead)
        return;
    is_dead = true;
    deleter();
  }

protected:
  static atexit_pfn_t deleter;
  static bool is_dead;
  static bool need_callback;

  static void atexit_callback()
  {
    need_callback = true;
    graceful_delete();
  }
};
template <class T> atexit_pfn_t DeletableSingleton<T>::deleter = 0;
template <class T> bool         DeletableSingleton<T>::is_dead = true;
template <class T> bool         DeletableSingleton<T>::need_callback = true;

///////////////////////////////////////////////////////////////////////
// Singleton
///////////////////////////////////////////////////////////////////////
template <typename T,
          template <class> class CreationPolicy = CreateUsingNew,
          template <class> class LifetimePolicy = DefaultLifetime
         >
class SingletonHolder
{
public:
    ///  Type of the singleton object
    typedef T object_type;

    ///  Returns a reference to singleton object
    static T& instance ();

private:
    // Helpers
    static void make_instance ();
    static void destroy_singleton ();

    // Protection
    SingletonHolder();

    // Data
    static T    *_p_instance;
    static bool  _destroyed;
};

template <class T,
          template <class> class C,
          template <class> class L
         >
T *SingletonHolder<T, C, L>::_p_instance = 0;

template <class T,
          template <class> class C,
          template <class> class L
         >
bool SingletonHolder<T, C, L>::_destroyed = false;

template <class T,
          template <class> class CreationPolicy,
          template <class> class LifetimePolicy
         >
inline T& SingletonHolder<T, CreationPolicy, LifetimePolicy>::instance()
{
  if (!_p_instance)
      make_instance ();
  return *_p_instance;
}

template <class T,
          template <class> class CreationPolicy,
          template <class> class LifetimePolicy
         >
void SingletonHolder<T, CreationPolicy, LifetimePolicy>::make_instance()
{
  if (!_p_instance)
    {
      if (_destroyed)
        {
          _destroyed = false;
          LifetimePolicy<T>::on_dead_reference();
        }
      _p_instance = CreationPolicy<T>::create();
      LifetimePolicy<T>::schedule_destruction (_p_instance, &destroy_singleton);
    }
}

template <class T,
          template <class> class CreationPolicy,
          template <class> class L
         >
void SingletonHolder<T, CreationPolicy, L>::destroy_singleton()
{
  // assert (!_destroyed);
  CreationPolicy<T>::destroy (_p_instance);
  _p_instance = 0;
  _destroyed = true;
}

#endif
