// reseau.cc

#include <vector>
#include <algorithm>

#include <cstdio>
#include <cstdlib>
#include <cmath>

#include "reseau.h"

static int NUMNODES[3];
/*****************************************************************************\
 * Reseau::init                                                              *
\*****************************************************************************/
void Reseau::init(int epsilon, int NMND[3])
{
  EPSILON = epsilon / 100.0;
  // Initialise NUMNODES
  NUMNODES[0] = NMND[0] = TAILLE_COUCHE_ENTREE;
  NUMNODES[1] = NMND[1] = TAILLE_COUCHE_CACHEE;
  NUMNODES[2] = NMND[2] = TAILLE_COUCHE_SORTIE;

  unit[0].resize(TAILLE_COUCHE_ENTREE);
  unit[1].resize(TAILLE_COUCHE_CACHEE);
  unit[2].resize(TAILLE_COUCHE_SORTIE);

  diff_sortie.resize(TAILLE_COUCHE_SORTIE);
  diff_cachee.resize(TAILLE_COUCHE_CACHEE);
  // Initialisation des trois couches de neurones
  for (int c = 0; c < 3; c++)
    for (int d = 0; d < NMND[c]; d++) // Pour chaque neurone de chaque couche...
    {
      unit[c][d].etat = 1;  // État initial à 1
      if (c != 0)
      {
        // Seuil aléatoire par défaut
        unit[c][d].seuil = (rand() % 100) / 100.0;

        // Initialisation des poids.
        for (int f = 0; f < NMND[c - 1]; f++) // Pour chaque neurone de la couche précédente
          unit[c][d].poids.push_back((rand() % 100) / 1000.0); // initialisation aléatoire du poids
      }
    }
}

/*****************************************************************************\
 * Reseau::set_epsilon                                                       *
\*****************************************************************************/
void Reseau::set_epsilon(int epsilon)
{
  EPSILON = epsilon / 100.0;
}

/*****************************************************************************\
 * Reseau::get_max_out_index                                                 *
\*****************************************************************************/
int Reseau::get_max_out_index() const
{
  for (int j = 0; j < TAILLE_COUCHE_SORTIE - 1; j++)
  {
    bool found = true;
    for (int i = j + 1; i < TAILLE_COUCHE_SORTIE;i++)
      if (unit[2][j].etat <= unit[2][i].etat)
      {
        found = false;
        break;
      }
    if (found)
      return j;
  }

  return TAILLE_COUCHE_SORTIE - 1;
}

/*****************************************************************************\
 * Reseau::copie_image_entree                                                *
\*****************************************************************************/
void Reseau::copie_image_entree(const float image[])
{
  for (int i = 0; i < TAILLE_COUCHE_ENTREE ; i++)
    unit[0][i].etat = image[i];
}

/*****************************************************************************\
 * Reseau::maj_activation                                                    *
\*****************************************************************************/
void Reseau::maj_activation(int couche)
{
  for (int i = 0; i < NUMNODES[couche]; i++)
  {
    float act;
    act = unit[couche][i].seuil;
    for (int v = 0; v < NUMNODES[couche - 1]; v++)
      act += unit[couche - 1][v].etat * unit[couche][i].poids[v];
    unit[couche][i].etat = 1.0 / (1.0 + exp(-act)); // fonction phi
  }
}

/*****************************************************************************\
 * Reseau::retro_propagation                                                 *
\*****************************************************************************/
void Reseau::retro_propagation(const float goal[])
{
  // Récupérations des erreurs sur les poids et les seuils.
  // calcul de dE/dy pour la couche de sortie
  for (int node = 0; node < TAILLE_COUCHE_SORTIE; node++)
  {
    float diff = goal[node] - unit[2][node].etat;
    diff_sortie[node] = diff * unit[2][node].etat * (1.0 - unit[2][node].etat);
  }
  // Calcul de dE/dy et de dE/ds pour la couche intermédiaire.
  // calcul de dE/dy
  for (int node = 0; node < TAILLE_COUCHE_CACHEE; node++)
  {
    float diff = 0;
    for (int tonode = 0; tonode < TAILLE_COUCHE_SORTIE; tonode++)
      diff += diff_sortie[tonode] * unit[2][tonode].poids[node];
    diff_cachee[node] = diff * unit[1][node].etat * (1.0 - unit[1][node].etat);
  }
  // Calcul des erreurs pour la couche cachée
  for (int neurone = 0; neurone < TAILLE_COUCHE_CACHEE; neurone++)
    unit[1][neurone].seuil += EPSILON * diff_cachee[neurone];
  for (int neurone = 0; neurone < TAILLE_COUCHE_CACHEE; neurone++)
    for (int inunit = 0; inunit < TAILLE_COUCHE_ENTREE; inunit++)
      unit[1][neurone].poids[inunit] += EPSILON * diff_cachee[neurone] * unit[0][inunit].etat;

  // Calcul des erreurs pour la couche de sortie
  for (int neurone = 0; neurone < TAILLE_COUCHE_SORTIE; neurone++)
    unit[2][neurone].seuil += EPSILON * diff_sortie[neurone];
  for (int neurone = 0; neurone < TAILLE_COUCHE_SORTIE; neurone++)
    for (int inunit = 0; inunit < TAILLE_COUCHE_CACHEE; inunit++)
      unit[2][neurone].poids[inunit] += EPSILON * diff_sortie[neurone] * unit[1][inunit].etat;
}
